//
// Copyright (c) 2000 by Tech Soft 3D, LLC.
// The information contained herein is confidential and proprietary to
// Tech Soft 3D, LLC., and considered a trade secret as defined under
// civil and criminal statutes.  Tech Soft 3D shall pursue its civil
// and criminal remedies in the event of unauthorized use or misappropriation
// of its trade secrets.  Use of this information by anyone other than
// authorized employees of Tech Soft 3D, LLC. is granted only under a
// written non-disclosure agreement, expressly prescribing the scope and
// manner of such use.
//
// $Header: /files/homes/master/cvs/hoops_master/docs_hoops/Hoops3DGS/prog_guide/examples/HUtility.cpp,v 1.2 2006-08-07 20:39:02 stage Exp $
//

// HUtility.cpp : implementation of the HUtility class
//


#include "HBaseView.h"
#include "HBaseOperator.h"
#include "HEventInfo.h"

#include "HUtility.h"
#include "HTools.h"
#include "hc.h"
#include "HStream.h"

#include "HConstantFrameRate.h"
#include "HUtilityGeometryCreation.h"


#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>


const float HUtility::PI=3.1415926f;
const float HUtility::EH_Epsilon=0.01f;

/////////////////////////////////////////////////////////////////////////////
// HOpConstruct
	

/*! 
  Order orders the input points in each dimension.  The resulting points are min and max for bounding box input.
  \param *one A pointer to an HPoint object.
  \param *two A pointer to an HPoint object.
*/
void HUtility::Order (HPoint *one, HPoint *two) {

	if (one->x > two->x) {
		float tmp;
		tmp = one->x; one->x = two->x; two->x = tmp;
	}
	if (one->y > two->y) {
		float tmp;
		tmp = one->y; one->y = two->y; two->y = tmp;
	}
	if (one->z > two->z) {
		float tmp;
		tmp = one->z; one->z = two->z; two->z = tmp;
	}
}


/*! 
  GetTime gets the system time in seconds.
*/
float HUtility::GetTime()
{
 	return  ((float)(clock())) /  (float) CLOCKS_PER_SEC;
}


/*! 
  InsertRectangle inserts a 2d rectangle as a HOOPS Polyline.
  \param *seg Segment name into which the rectangle should be inserted.
  \param x0 X-value of vertex 1 of the rectangle.
  \param y0 Y-value of vertex 1 of the rectangle.
  \param x1 X-value of vertex 3 of the rectangle.
  \param y1 Y-value of vertex 3 of the rectangle.
*/
void HUtility::InsertRectangle (
   const char	*	seg,
   float       x0,
   float       y0,
   float       x1,
   float       y1) {
   HPoint       pts[5];

   pts[0].x = x0;   pts[0].y = y0;   pts[0].z = 0.0F;
   pts[1].x = x0;   pts[1].y = y1;   pts[1].z = 0.0F;
   pts[2].x = x1;   pts[2].y = y1;   pts[2].z = 0.0F;
   pts[3].x = x1;   pts[3].y = y0;   pts[3].z = 0.0F;
   pts[4].x = x0;   pts[4].y = y0;   pts[4].z = 0.0F;

   HC_QInsert_Polyline (seg, 5, pts);
}


/*! 
  PointsEqual compares the equality of two (x,y,z) points.
  \param *point1 A pointer to an HPoint object.
  \param *point2 A pointer to an HPoint object.
  \return A Boolean value indicating equality or inequality
*/
bool HUtility::PointsEqual(HPoint *point1, HPoint *point2) {

	if (feq(point1->x, point2->x, 0.0001) &&
		feq(point1->y, point2->y, 0.0001) &&
		feq(point1->z, point2->z, 0.0001))
		return true;
	else
		return false;
}

/*! 
  ClampPointToWindow clamps a point to its nearest corner in window space.
  \param *point A pointer to an HPoint object.
*/
void HUtility::ClampPointToWindow (HPoint *point) {

    if      (point->x < -1.0)  point->x = -1.0F;
    else if (point->x >  1.0)  point->x =  1.0F;

    if      (point->y < -1.0)  point->y = -1.0F;
    else if (point->y >  1.0)  point->y =  1.0F;

    if      (point->z < -1.0)  point->z = -1.0F;
    else if (point->z >  1.0)  point->z =  1.0F;


    return;

}


/*! 
  GenerateCuboidPoints creates the vertices of a cuboid based on two input points.
  \param *max A pointer to an HPoint object.
  \param *min A pointer to an HPoint object.
  \param *points A pointer to an array of HPoint objects.  Passed by reference.
*/
void HUtility::GenerateCuboidPoints(    
	HPoint	*max,
    HPoint	*min,
	HPoint	*points){

    points[0].x = min->x;
    points[0].y = min->y;
    points[0].z = min->z;

    points[1].x = max->x;
    points[1].y = min->y;
    points[1].z = min->z;

    points[2].x = max->x;
    points[2].y = max->y;
    points[2].z = min->z;

    points[3].x = min->x;
    points[3].y = max->y;
    points[3].z = min->z;

    points[4].x = min->x;
    points[4].y = min->y;
    points[4].z = max->z;
    
    points[5].x = max->x;
    points[5].y = min->y;
    points[5].z = max->z;

    points[6].x = max->x;
    points[6].y = max->y;
    points[6].z = max->z;

    points[7].x = min->x;
    points[7].y = max->y;
    points[7].z = max->z;
}


/*! 
  InsertBox inserts a six-sided, rectangular cuboid based on 2 points.
  \param *max A pointer to an HPoint object.
  \param *min A pointer to an HPoint object.
*/
void HUtility::InsertBox (
    HPoint	*max,
    HPoint	*min){
    HPoint 	points[8];
    int		face_list[30];

	GenerateCuboidPoints(max, min, points);

    face_list[0] = 4;
    face_list[1] = 0;
    face_list[2] = 1;
    face_list[3] = 2;
    face_list[4] = 3;
    
    face_list[5] = 4;
    face_list[6] = 7;
    face_list[7] = 6;
    face_list[8] = 5;
    face_list[9] = 4;
    
    face_list[10] = 4;
    face_list[11] = 1;
    face_list[12] = 5;
    face_list[13] = 6;
    face_list[14] = 2;
    
    face_list[15] = 4;
    face_list[16] = 4;
    face_list[17] = 0;
    face_list[18] = 3;
    face_list[19] = 7;
    
    face_list[20] = 4;
    face_list[21] = 3;
    face_list[22] = 2;
    face_list[23] = 6;
    face_list[24] = 7;
    
    face_list[25] = 4;
    face_list[26] = 4;
    face_list[27] = 5;
    face_list[28] = 1;
    face_list[29] = 0;

	HC_Insert_Shell (8, points, 30, face_list);
   
    return;
}


/*! 
  CreateAxisCircle creates a circle based on a centerpoint and a radius. Note this function actually inserts a circle into
  the HOOPS scene using the HOOPS routine HC_Insert_Circle, whereas GeneratePointsOnCircle only generates an array of 
  points along a circular trajectory.
  \param center The center of the desired circle.
  \param radius The length of the radius of the circle.
  \param axis XY, XZ, or YZ. This defines the view plane in which the circle will be drawn.
  \return A Boolean value indicating success or failure.
*/
bool HUtility::CreateAxisCircle (HPoint & center,float radius,int axis)
{
	HPoint points[3];
	bool success;

	// create three points on the circle in the appropriate view plane

	success = HUtilityGeometryCreation::GeneratePointsOnCircle(points, center, radius, axis, 3);
	if (!success)
		return success;

	HC_Insert_Circle(&points[0], &points[1], &points[2]);
	return true;

}


/*! 
  NormalizeDoublePrecisionVector normalizes a double-precision vector  
  \param *vector A pointer to a double precision vector.
*/
void HUtility::NormalizeDoublePrecisionVector(double* vector)
{
	double length;
	length = sqrt(vector[0]*vector[0]+vector[1]*vector[1]+vector[2]*vector[2]);
	vector[0]/=length;
	vector[1]/=length;
	vector[2]/=length;
}


/*! 
  ComputeDoublePrecisionVectorLength computes the length of a double precision vector.
  \param *vector A pointer to a double precision vector
  \return The double precision length of the vector.
*/
double HUtility::ComputeDoublePrecisionVectorLength( double * vector )
{
	return sqrt(vector[0]*vector[0]+vector[1]*vector[1]+vector[2]*vector[2]);
}


/*! 
  ComputeDoublePrecisionCrossProduct computes the cross product of two double precision vectors.
  \param *vector1 A pointer to a double precision vector
  \param *vector2 A pointer to a double precision vector
  \param *up A pointer to a double precision cross product.  Passed by reference.
*/
void HUtility::ComputeDoublePrecisionCrossProduct(double * vector1, double * vector2, double * up)
{
	up[0] = vector1[1] * vector2[2]  -  vector1[2] * vector2[1];
    up[1] = vector1[2] * vector2[0]  -  vector1[0] * vector2[2];
    up[2] = vector1[0] * vector2[1]  -  vector1[1] * vector2[0];
}


/*! 
  ComputeDoublePrecisionDotProduct computes a double precision dot product of two vectors.
  \param *vector1 A pointer to a double precision vector.
  \param *vector2 A pointer to a double precision vector.
  \return The double precision dot product of the vectors.
*/
double HUtility::ComputeDoublePrecisionDotProduct(double * vector1, double * vector2)
{
	return (vector1[0] * vector2[0]) + (vector1[1] * vector2[1]) + (vector1[2] * vector2[2]);
}


// 
/*! 
  GetViewplaneVectors computes the construction-plane-constrained vectors used for geometry insertion.
  \param seg_key Key to the construction segment
  \param *view A pointer to the double precision view vector.  Passed by reference.
  \param *up A pointer to the double precision up vector. Passed by reference.
*/
void HUtility::GetViewplaneVectors(HC_KEY seg_key, double * view, double * up)
{
	HPoint	position, target;
	HVector	up_vector, view_vector, orthogonal;
	float	x, y, z;
	double	ortho[3];

	HC_Open_Segment_By_Key(seg_key);

		HC_Show_Net_Camera_Up_Vector(&x, &y, &z);
		up_vector.Set(x, y, z);

		HC_Show_Net_Camera_Position (&x, &y, &z);
		position.Set(x, y, z);

		HC_Show_Net_Camera_Target (&x, &y, &z);
		target.Set(x, y, z);

	HC_Close_Segment();

	view_vector.Set(position.x - target.x,
					position.y - target.y,
					position.z - target.z);

	// find vector in plane orthogonal to view and up plane
	HC_Compute_Cross_Product(&view_vector, &up_vector, &orthogonal);

	// now do the casting to double
	view[0] = (double) position.x - (double) target.x;
	view[1] = (double) position.y - (double) target.y;
	view[2] = (double) position.z - (double) target.z;

	NormalizeDoublePrecisionVector(view);

	ortho[0] = (double) orthogonal.x;
	ortho[1] = (double) orthogonal.y;
	ortho[2] = (double) orthogonal.z;

	NormalizeDoublePrecisionVector(ortho);

	ComputeDoublePrecisionCrossProduct(view, ortho, up);
}






/*! 
  InsertArrow inserts an arrow consisting of a cylinder and a cone at the origin. This function is used by the Axis 
  display.  The arrow is completely customizable.
  \param stick_length Length of the cylindrical shaft of the arrow.
  \param tip_length Length of the conical tip of the arrow.
  \param stick_radius Radius of the cylindrical shaft of the arrow.
  \param tip_radius Radius of the circular base of the arrow's conical tip.
  \param *text Pointer to a string that annotates the arrow.
  \param *stick_color Color of the shaft of the arrow.
  \param *tip_color Color of the tip of the arrow.
  \param *num_sides Number of sides of arrow (default = 6)
  \param tip_falloff Radius reduction value on stick (default = 0.05)
*/
void HUtility::InsertArrow(float stick_length,float tip_length, float stick_radius, float tip_radius, const char *text, const char *stick_color, const char *tip_color, int num_sides, float tip_falloff)
{
  	HPoint	bottom, top;	

  	HPoint *stick_points = 0;
  	HPoint *tip_points = 0;
	int *stick_faces_list = 0;
	int *tip_faces_list = 0;
	int i;

	stick_points= new HPoint[num_sides*2];				
	tip_points= new HPoint[num_sides+1];
  	stick_faces_list= new int[num_sides*5+num_sides+1];
  	tip_faces_list= new int[num_sides*4];

	bottom.Set(0.0,0.0,0.0);
	top.Set(0.0,stick_length,0.0);
	
	tip_points[num_sides].Set(0.0,top.y+tip_length,0.0);
  
	HUtilityGeometryCreation::GeneratePointsOnCircle(stick_points, bottom, stick_radius-tip_falloff, HUtility::XZ, num_sides,false);
	HUtilityGeometryCreation::GeneratePointsOnCircle(stick_points+num_sides, top, stick_radius, HUtility::XZ, num_sides,false);
	HUtilityGeometryCreation::GeneratePointsOnCircle(tip_points, top, tip_radius, HUtility::XZ, num_sides,false);
  
	for (i=0;i<num_sides;i++)
	{
		 stick_faces_list[i*5]=4;
		 stick_faces_list[i*5+1]=i;
		 stick_faces_list[i*5+2]=i+num_sides;
		 stick_faces_list[i*5+3]=((i+1)%num_sides)+num_sides;
		 stick_faces_list[i*5+4]=(i+1)%num_sides;
		 
		 tip_faces_list[i*4]=3;
		 tip_faces_list[i*4+1]=i;
		 tip_faces_list[i*4+2]=(i+1)%num_sides;
		 tip_faces_list[i*4+3]=num_sides;

		 stick_faces_list[num_sides*5+i+1]=i;		//this constructs the face for the bottom of
													//the cylinder (not really needed for the axis)
	}
	stick_faces_list[num_sides*5]=num_sides;

	 
	if (stick_color)
	{
	    HC_Open_Segment("");
		HC_Set_Color(stick_color);
		HC_Insert_Shell(num_sides*2,stick_points,num_sides*5+num_sides+1,stick_faces_list);
	    HC_Close_Segment();

	}else
	    HC_Insert_Shell(num_sides*2,stick_points,num_sides*5+num_sides+1,stick_faces_list);

	
	if (text)	
	{
 	    HC_Set_Text_Font("name=stroked,size=0.2oru, transforms=character position only");
	    HC_Insert_Text(0,top.y+tip_length+0.1,0.0,text);
	    HC_Set_Color("text = white");
	}
 
	if (tip_color)
	{
	    HC_Open_Segment("");	
		HC_Set_Color(tip_color);
 		HC_Insert_Shell(num_sides+1,tip_points,num_sides*4,tip_faces_list);
 	    HC_Close_Segment();
	}else
	    HC_Insert_Shell(num_sides+1,tip_points,num_sides*4,tip_faces_list);


	if (stick_points)
		delete stick_points;
	if (tip_points)
		delete tip_points;
	if (stick_faces_list)
		delete stick_faces_list;
	if (tip_faces_list)
		delete tip_faces_list;
}

/*! 
   Rotates the contents of an opened segment to a given vector in space. 
  \param start First point in space that defines vector
  \param end Second point in space that defines vector
*/

void HUtility::RotateToVector(HPoint start, HPoint end)
{
	
	HPoint gl,gl2;
	HPoint a,b;
	float degreeangley, degreeanglez;
	gl.Set(start.x-end.x, start.y-end.y,start.z-end.z);
 	//float vlen=(float)HC_Compute_Vector_Length(&gl);
	gl2.Set(gl.x,0,gl.z);
	a.Set(gl.x,gl.y,gl.z);
	b.Set(gl.x,0,gl.z);


	float a_squared = a.x*a.x+a.y*a.y+a.z*a.z;
	float b_squared = b.x*b.x+b.y*b.y+b.z*b.z;

	if (a_squared == 0)
		a_squared =1;

	if (b_squared == 0)
		b_squared =1;


 	if (start.x == end.x && start.z == end.z && start.z == end.x && start.x == end.z)
		degreeangley = 90;
	else
		degreeangley = (float)acos((a.x*b.x + a.y*b.y + a.z*b.z) / sqrt(a_squared*b_squared))* (180.0f/3.1415927f);
  	a.Set(start.x-end.x,0,start.z-end.z);
	b.Set(start.x-end.x,0,0);
 
	if (start.x == end.x && start.y == end.y && start.y == end.x && start.y == end.x)
		degreeanglez = 90;
	else
		degreeanglez = (float)acos((a.x*b.x + a.y*b.y + a.z*b.z) / sqrt(a_squared*b_squared)) * (180.0f/3.1415927f);
	
	if (end.x<=start.x && end.z<=start.z)	
		;
	else if (end.x<=start.x && end.z>start.z)	
		degreeanglez=-degreeanglez;
	else if (end.x>start.x && end.z>start.z)	
		degreeanglez=-(180-degreeanglez);
	else 	
		degreeanglez=(180-degreeanglez);	
	if (end.y>start.y)
		degreeangley=-degreeangley;		 
	HC_Rotate_Object(-90.0,0.0,0.0);
	HC_Rotate_Object(-degreeangley,0,0);
	HC_Rotate_Object(0.0,-(-(90-degreeanglez)),0);				
	HC_Translate_Object(start.x,start.y,start.z);
	
}

/*
  {axis.x, axis.y, axis.z, 0, 
  ref.x, ref.y, ref.z, 0,
  side.x, side.y, side.z, 0,
  pos.x, pos.y, pos.z,1};

where:
  axis:  viewing vector (i.e. the camera target -camera position vector) 
  ref:    camera up vector
  side: the cross product of the axis and ref (i.e. orthogonal to the axis and ref vectors) 
  pos:  camera target


*/

void HUtility::GetCameraFromMatrix(float *matrix, HPoint *target, HPoint *position, HPoint *up_vector, HPoint *translation)
{
//Target remains unchanged
//    HPoint view_vector;
//    view_vector.Set(position.x-target.x, position.y-target.y, position.z-target.z);

    position->x = matrix[0];
    position->y = matrix[1];
    position->z = matrix[2];

    up_vector->x = matrix[4];
    up_vector->y = matrix[5];
    up_vector->z = matrix[6];

    translation->x = matrix[12];
    translation->y = matrix[13];
    translation->z = matrix[14];
}

#if 0
static void transpose(float *in, float *out)
{
    float temp[16];

    temp[0] = in[0];
    temp[1] = in[4];
    temp[2] = in[8];
    temp[3] = in[12];
    temp[4] = in[1];
    temp[5] = in[5];
    temp[6] = in[9];
    temp[7] = in[13];
    temp[8] = in[2];
    temp[9] = in[6];
    temp[10] = in[10];
    temp[11] = in[14];
    temp[12] = in[3];
    temp[13] = in[7];
    temp[14] = in[11];
    temp[15] = in[15];

    memcpy(out, temp, 16*sizeof(float));

    return;
}
#endif



void HUtility::GetMatrixFromCamera(HPoint target, HPoint position, HPoint up_vector, HPoint translation, float *matrix)
{
    HPoint view_vector;
    view_vector.Set(target.x-position.x, target.y-position.y, target.z-position.z);

    HC_Compute_Normalized_Vector(&view_vector, &view_vector);
    HC_Compute_Normalized_Vector(&up_vector, &up_vector);

    HPoint cross1;

    HC_Compute_Cross_Product(&up_vector, &view_vector, &cross1);

    matrix[0] = cross1.x;
    matrix[1] = cross1.y;
    matrix[2] = cross1.z;
    matrix[3] = 0.0f;

    HPoint cross2;

    HC_Compute_Cross_Product(&view_vector, &cross1, &cross2);

    matrix[4] = cross2.x;
    matrix[5] = cross2.y;
    matrix[6] = cross2.z;
    matrix[7] = 0.0f;

    matrix[8] = view_vector.x;
    matrix[9] = view_vector.y;
    matrix[10] = view_vector.z;
    matrix[11] = 0.0f;
    
    matrix[12] = translation.x;
    matrix[13] = translation.y;
    matrix[14] = translation.z;
    matrix[15] = 1.0f;
}


void HUtility::SmoothTransition(HPoint &co, HPoint &to, HPoint &uo, float fieldwidtho, float fieldheighto, HPoint &cn, HPoint &tn, HPoint &un, float fieldwidthn, float fieldheightn, HBaseView *m_pHView)
{
	if (m_pHView->GetRenderMode()==HRenderHiddenLine)
	{
 	    HC_Set_Camera_Target(tn.x,tn.y,tn.z);
 	    HC_Set_Camera_Position(cn.x,cn.y,cn.z);
 	    HC_Set_Camera_Up_Vector(un.x,un.y,un.z);
		HC_Set_Camera_Field(fieldwidthn,fieldheightn);

 	    m_pHView->CameraPositionChanged();
	    m_pHView->Update();
	    return;
	}


	bool vis=false;
	char vis_settings[1024];
	
	if(HC_Show_Existence("visibility"))
	{
		HC_Show_Visibility(vis_settings);
		vis=true;
	}

	HC_Set_Visibility("text=off");
	HC_Set_Rendering_Options("attribute lock=(visibility=(text))");
	
	float matrix_old[16],  matrix_new[16];
	HPoint view_vector_old, view_vector_new;

	view_vector_old.Set(to.x - co.x, to.y - co.y, to.z - co.z);
	view_vector_new.Set(tn.x - cn.x, tn.y - cn.y, tn.z - cn.z);


	float length_delta = (float)HC_Compute_Vector_Length(&view_vector_new) - (float)HC_Compute_Vector_Length(&view_vector_old);
  
	HC_Compute_Normalized_Vector(&view_vector_old,&view_vector_old);
	HC_Compute_Normalized_Vector(&view_vector_new,&view_vector_new);

	bool need_rotation=true;
	HPoint rotation_delta;

	rotation_delta.Set(view_vector_old.x - view_vector_new.x,
	    view_vector_old.y - view_vector_new.y,
	    view_vector_old.z - view_vector_new.z);

	if ((float)HC_Compute_Vector_Length(&rotation_delta) < 0.0001f)
	    need_rotation=false;

	// normalise the vectors
	float length = (float)HC_Compute_Vector_Length(&co);

	HPoint translation;
	translation.Set(0.0f,0.0f,0.0f);

	HPoint position_move_delta;

	if (need_rotation)
	{
	    HUtility::GetMatrixFromCamera(to, co, uo, translation, matrix_old);
	    HUtility::GetMatrixFromCamera(tn, cn, un, translation, matrix_new);
	}
	else
	{
	    position_move_delta.Set(cn.x-co.x, cn.y-co.y, cn.z-co.z);
	}


	float	matrix_interim[16];
	float	quat_old[4], quat_new[4], quat_interim[4];

	if (need_rotation)
	{
	    HUtility::MatrixToQuaternion(matrix_old, quat_old);
	    HUtility::MatrixToQuaternion(matrix_new, quat_new);
	}

	HPoint target_move_delta;
	target_move_delta.Set(tn.x-to.x, tn.y-to.y, tn.z-to.z);

	float field_width_delta = fieldwidthn - fieldwidtho;
	float field_height_delta = fieldheightn - fieldheighto;

	float dot1, dot2;

	if (need_rotation)
	{
	    // Simple dot product to determine which direction the quaternion should 
	    // go if we want to take the shortest path.
	    dot1 = ((quat_old[0] - quat_new[0])*(quat_old[0] - quat_new[0])
			+ (quat_old[1] - quat_new[1])*(quat_old[1] - quat_new[1])
			+ (quat_old[2] - quat_new[2])*(quat_old[2] - quat_new[2])
			+ (quat_old[3] - quat_new[3])*(quat_old[3] - quat_new[3])); 

	    dot2 = ((quat_old[0] + quat_new[0])*(quat_old[0] + quat_new[0])
			+ (quat_old[1] + quat_new[1])*(quat_old[1] + quat_new[1])
			+ (quat_old[2] + quat_new[2])*(quat_old[2] + quat_new[2])
			+ (quat_old[3] + quat_new[3])*(quat_old[3] + quat_new[3])); 

	    if (dot2 < dot1) {
		    quat_new[0] = -quat_new[0];
		    quat_new[1] = -quat_new[1];
		    quat_new[2] = -quat_new[2];
		    quat_new[3] = -quat_new[3];
	    }
	}

	HPoint position, up_vector;
 
	float total_duration = 0.5;							
	float initial_delta = (float) 0.05 / total_duration;			//0.05 = 20fps	(guessing)
	float q, total_time_start, current_time;
	
	HC_Show_Time(&total_time_start);
    
	HC_Open_Segment_By_Key(m_pHView->GetSceneKey());
	
	do
	{
		HC_Show_Time(&current_time);		
		q =  ((current_time - total_time_start) / total_duration) + initial_delta;
		if (q>=1.0)
		{
			q = 1.0;
			HC_Set_Camera_Target(tn.x,tn.y,tn.z);
 			HC_Set_Camera_Position(cn.x,cn.y,cn.z);
 			HC_Set_Camera_Up_Vector(un.x,un.y,un.z);
		}
		else
		{
			if (need_rotation)
			{
				// first let use quaternions to take care of the interpolation 
				// of the rotation element of the camera movement
				HUtility::TransitionQuaternion(quat_old, quat_new, q, quat_interim);
				HUtility::QuaternionToMatrix(quat_interim, matrix_interim);
				
				position.x = to.x - (matrix_interim[8] * (length+length_delta*q));
				position.y = to.y - (matrix_interim[9] * (length+length_delta*q));
				position.z = to.z - (matrix_interim[10] * (length+length_delta*q));
				
				up_vector.x = matrix_interim[4];
				up_vector.y = matrix_interim[5];
				up_vector.z = matrix_interim[6];
				
				HC_Set_Camera_Position(position.x, position.y, position.z);
				HC_Set_Camera_Up_Vector(up_vector.x, up_vector.y, up_vector.z);
			}
			else
			{
				HC_Set_Camera_Position(co.x + (q*position_move_delta.x), 
					co.y + (q*position_move_delta.y), 
					co.z + (q*position_move_delta.z));
			}
			
			// now linearly interpolate the camera target and field changes
			HC_Set_Camera_Target(to.x + (q*target_move_delta.x), 
				to.y + (q*target_move_delta.y), 
				to.z + (q*target_move_delta.z));
		}
		
		HC_Set_Camera_Field(fieldwidtho+(q*field_width_delta), 
			fieldheighto+(q*field_height_delta));
		
		m_pHView->CameraPositionChanged();
		m_pHView->Update();
	}while (q < 1.0);
    HC_Close_Segment();
	
	HC_UnSet_One_Rendering_Option("attribute lock");
	HC_UnSet_One_Visibility("text");
	
	if(vis)
		HC_Set_Visibility(vis_settings);

	m_pHView->Update();
}


#if 0
void HUtility::SmoothTransition(HPoint &co,HPoint &to, HPoint &uo, HPoint &cn, HPoint &tn, HPoint &un, HBaseView *m_pHView)
{
	double total_duration = 1.0;
	double framerate = 0.01f;
	double distance = 0.0;
	double total_distance = 0.0;
	double duration = 0.0;

	if (!m_pHView->GetSmoothTransition() || m_pHView->GetRenderMode()==HRenderHiddenLine)
	{

 		HC_Set_Camera_Target(tn.x,tn.y,tn.z);
 		HC_Set_Camera_Position(cn.x,cn.y,cn.z);
 		HC_Set_Camera_Up_Vector(un.x,un.y,un.z);
 		HC_Move_Distant_Light(m_pHView->GetDefaultLightKey(),
                              cn.x,cn.y,cn.z);
 
		m_pHView->Update();
		return;
	}

	// smooth transition modifies the location of the camera rather than simply changing the camera field.
	// consequently it does not have any effect if we're in perspective mode.  to workaround this we change
	// to perspective mode temporaily as we do the smooth transition
	bool orthographicmode = false;

	if (m_pHView->GetProjMode() != ProjPerspective)
	{
		m_pHView->SetProjMode(ProjPerspective);
		orthographicmode = true;
	}
	
	HPoint vt,vc,vu;
	vt.Set(tn.x-to.x,tn.y-to.y,tn.z-to.z);
	vc.Set(cn.x-co.x,cn.y-co.y,cn.z-co.z);
	vu.Set(un.x-uo.x,un.y-uo.y,un.z-uo.z);

	HConstantFrameRate *cframerate = m_pHView->GetConstantFrameRateObject();
 	float total_time_start;
	bool cfr_active = cframerate->GetActive();
	if (!cfr_active)
		cframerate->Start();

	HC_Show_Time(&total_time_start);
	while (1)
	{
		distance = (1.0-distance)/(total_duration-duration)*framerate;
		if (distance<0)
			distance=0.1;
		total_distance +=distance;
		if (total_distance>1.0)
			total_distance = 1.0;
 		HC_Set_Camera_Target(to.x+vt.x*total_distance,to.y+vt.y*total_distance,to.z+vt.z*total_distance);
		HC_Set_Camera_Position(co.x+vc.x*total_distance,co.y+vc.y*total_distance,co.z+vc.z*total_distance);
		HC_Set_Camera_Up_Vector(uo.x+vu.x*total_distance,uo.y+vu.y*total_distance,uo.z+vu.z*total_distance);

		m_pHView->CameraPositionChanged();

 		float update_time_start,update_time_end;
		HC_Show_Time(&update_time_start);
		m_pHView->Update();	
		HC_Show_Time(&update_time_end);
		framerate = update_time_end-update_time_start;
		if (framerate<0.01)
		{
			float pausetime_start;
			HC_Show_Time(&pausetime_start);
			while (1)
			{
				float pausetime_end;
				HC_Show_Time(&pausetime_end);
				if (pausetime_end-pausetime_start>(0.01-framerate))
					break;
			}
 			framerate = 0.01f;
			HC_Show_Time(&update_time_end);

		}
		duration = update_time_end - total_time_start;
//saftey measure if infinite loop problem occurs (should never happen!!!)
		if (duration>total_duration+total_duration/4)
		{
			HC_Set_Camera_Target(tn.x,tn.y,tn.z);
			HC_Set_Camera_Position(cn.x,cn.y,cn.z);
			HC_Set_Camera_Up_Vector(un.x,un.y,un.z);
 			//HC_Move_Distant_Light(m_pHView->GetDefaultLightKey(),
	          //                    cn.x,cn.y,cn.z);
			m_pHView->CameraPositionChanged();
			m_pHView->Update();
			break;
		}


		if (total_distance>=1.0)
			break;

	}
	if (!cfr_active)
		cframerate->Stop();

	if (orthographicmode)
	{
		m_pHView->SetProjMode(ProjOrthographic);
		m_pHView->Update();
	}
}
#endif 

void HUtility::ComputeVectorToVectorRotationMatrix(float *matrix,HPoint p1,HPoint p2)
{
    HPoint p3;

    HC_Compute_Normalized_Vector(&p1,&p1);
    HC_Compute_Normalized_Vector(&p2,&p2);
    HC_Compute_Cross_Product(&p1,&p2,&p3);

    float dprod = (float)HC_Compute_Dot_Product(&p1,&p2);
    float l=(float)HC_Compute_Vector_Length(&p3);
    //float ang2 = (float)asin(l);
    float ang = (float)atan2(l,dprod);
    ang*=(180.0f/(float)PI);

    if (l==0)
    {
	//we need to handle the anti-parallel case
        HC_Compute_Identity_Matrix(matrix);
    }
    else
        HC_Compute_Offaxis_Rotation(p3.x,p3.y,p3.z,-ang, matrix);
}
    	
bool HUtility::ExtractRGB(char *text,float &r, float &g, float &b)
{
	int l= (int) strlen(text);
	for(int i=0;i<l;i++) { text[i]=(char)toupper((int)text[i]); }
	char *rpos = strstr(text,"R=");
	if (!rpos)
		return false;
	r=(float)atof(rpos+2);
	rpos = strstr(text,"G=");
	g=(float)atof(rpos+2);
	rpos = strstr(text,"B=");
	b=(float)atof(rpos+2);
	return true;
}

#define X 0
#define Y 1
#define Z 2
#define W 3

void HUtility::QuaternionToMatrix(float * quaternion, float *matrix)
{	
	double s, xs, ys, zs, wx, wy, wz, xx, xy, xz, yy, yz, zz;

	s = 2.0/(quaternion[X]*quaternion[X] + quaternion[Y]*quaternion[Y] + quaternion[Z]*quaternion[Z] + quaternion[W]*quaternion[W]);

	xs = quaternion[X]*s;	ys = quaternion[Y]*s;		zs = quaternion[Z]*s;
	wx = quaternion[W]*xs;	wy = quaternion[W]*ys;		wz = quaternion[W]*zs;
	xx = quaternion[X]*xs;	xy = quaternion[X]*ys;		xz = quaternion[X]*zs;
	yy = quaternion[Y]*ys;	yz = quaternion[Y]*zs;		zz = quaternion[Z]*zs;

	matrix[0] = 1.0f - (float)(yy+zz);
	matrix[1] = (float)xy + (float)wz;
	matrix[2] = (float)xz - (float)wy;
	matrix[3] = 0.0f;

	matrix[4] = (float)xy - (float)wz;
	matrix[5] = 1.0f - (float)(xx + zz);
	matrix[6] = (float)yz + (float)wx;
	matrix[7] = 0.0f;

	matrix[8] = (float)xz + (float)wy;
	matrix[9] = (float)yz - (float)wx;
	matrix[10] = 1.0f - (float)(xx + yy);
	matrix[11] = 0.0f;

	matrix[12] = 0.0f;
	matrix[13] = 0.0f;
	matrix[14] = 0.0f;
	matrix[15] = 1.0f;
}

void HUtility::MatrixToQuaternion(float *matrix, float *quaternion)
{
	double	tr, s;
	int		i, j, k;

	tr = matrix[0] + matrix[5] + matrix[10];
	if (tr > 0.0)
	{
		s = sqrt(tr + 1.0);
		quaternion[W] = (float)s * 0.5f;
		s = 0.5f/(float)s;

		quaternion[X] = (matrix[6] - matrix[9]) * (float)s;
		quaternion[Y] = (matrix[8] - matrix[2]) * (float)s;
		quaternion[Z] = (matrix[1] - matrix[4]) * (float)s;
	}
	else
	{
		i = X;
		if (matrix[(Y*4)+Y] > matrix[(X*4)+X])
			i = Y;
		if (matrix[(Z*4)+Z] > matrix[(i*4)+i])
			i = Z;

		int nxt[3] = {Y,Z,X};

		j = nxt[i];
		k = nxt[j];

		s = sqrt(matrix[(i*4)+i] - (matrix[(j*4)+j] + matrix[(k*4)+k]) + 1.0f);

		quaternion[i] = (float)s * 0.5f;
		s = 0.5f / s;

		quaternion[W] = (matrix[(j*4) + k] - matrix[(k*4) + j]) * (float)s;
		quaternion[j] = (matrix[(i*4) + j] + matrix[(j*4) + i]) * (float)s;
		quaternion[k] = (matrix[(i*4) + k] + matrix[(k*4) + i]) * (float)s;
	}
}

void HUtility::TransitionQuaternion(float *quat_old, float *quat_new, float t, float *qt)
{
	double omega, cosom, sinom, sclp, sclq;

	cosom = quat_old[X]*quat_new[X] + quat_old[Y]*quat_new[Y] + quat_old[Z]*quat_new[Z] + quat_old[W]*quat_new[W];

	if ((1.0 + cosom) > EPSILON)
	{
		if ((1.0 - cosom) > EPSILON)
		{
			omega = acos(cosom);
			sinom = sin(omega);
			sclp = sin((1.0 - t)*omega)/sinom;
			sclq = sin(t*omega)/sinom;
		}
		else
		{
			sclp = 1.0 - t;
			sclq = t;
		}

		for (int i=0; i<4; i++)
			qt[i] = (float)sclp*quat_old[i] +(float)sclq*quat_new[i];
	}
	else
	{
		qt[X] = -quat_old[Y];	qt[Y] = quat_old[X];
		qt[Z] = -quat_old[W];	qt[W] = quat_old[Z];
		sclp = sin((1.0 -t) * PI/2);
		sclq = sin(t * PI/2);
		for (int i=0; i<3; i++)
			qt[i] = (float)sclp*quat_old[i] + (float)sclq*qt[i];
	}
}


/*! 
   Animates the rotation of an object from the object space of one segment to the object space of another.  Uses quaternions to 
   make the transition appear smmoth
  \param m_pView Pointer to the HBaseView object
  \param old_seg Segment from which the transition should occur.
  \param new_seg Segment which has the modelling matrix which we want to transition to
  \param num_of_transitions numer of frames in the animation
*/

void HUtility::SmoothQuaternionTransition(HBaseView *m_pView, char *old_seg, char *new_seg, int num_of_transitions)
{
    float matrix_old[16], matrix_new[16];

    HC_Open_Segment(new_seg);
	HC_Show_Modelling_Matrix(matrix_new);
	matrix_new[12] = 0;
	matrix_new[13] = 0;
	matrix_new[14] = 0;
	matrix_new[15] = 1;
    HC_Close_Segment();

    // need to determine the scaling factor as the quaternions do not take scales into account
    float new_scale = (float) sqrt((matrix_new[0]*matrix_new[0])
				+(matrix_new[1]*matrix_new[1])
				+(matrix_new[2]*matrix_new[2]));

    // need to remove the scaling factor from the matrix before we turn it into a quaternion
    matrix_new[0] /= new_scale;
    matrix_new[1] /= new_scale;
    matrix_new[2] /= new_scale;
    matrix_new[4] /= new_scale;
    matrix_new[5] /= new_scale;
    matrix_new[6] /= new_scale;
    matrix_new[8] /= new_scale;
    matrix_new[9] /= new_scale;
    matrix_new[10] /= new_scale;

    // now get the old segments matrix.  translation has already been removed
    HC_Open_Segment(old_seg);
	HC_Show_Modelling_Matrix(matrix_old);
    HC_Close_Segment();

    // Again we need to determine and remove the scaling factor from the matrix
    float old_scale = (float) sqrt((matrix_old[0]*matrix_old[0])
				    +(matrix_old[1]*matrix_old[1])
				    +(matrix_old[2]*matrix_old[2]));

    matrix_old[0] /= old_scale;
    matrix_old[1] /= old_scale;
    matrix_old[2] /= old_scale;
    matrix_old[4] /= old_scale;
    matrix_old[5] /= old_scale;
    matrix_old[6] /= old_scale;
    matrix_old[8] /= old_scale;
    matrix_old[9] /= old_scale;
    matrix_old[10] /= old_scale;

    // now turn the matrices into quaternions
    float	matrix_interim[16];
    float	quat_old[4], quat_new[4], quat_interim[4];

    HUtility::MatrixToQuaternion(matrix_old, quat_old);
    HUtility::MatrixToQuaternion(matrix_new, quat_new);

    // Simple dot product to determine which direction the quaternion should 
    // go if we want to take the shortest path.
    float dot1 = ((quat_old[0] - quat_new[0])*(quat_old[0] - quat_new[0])
		    + (quat_old[1] - quat_new[1])*(quat_old[1] - quat_new[1])
		    + (quat_old[2] - quat_new[2])*(quat_old[2] - quat_new[2])
		    + (quat_old[3] - quat_new[3])*(quat_old[3] - quat_new[3])); 

    float dot2 = ((quat_old[0] + quat_new[0])*(quat_old[0] + quat_new[0])
		    + (quat_old[1] + quat_new[1])*(quat_old[1] + quat_new[1])
		    + (quat_old[2] + quat_new[2])*(quat_old[2] + quat_new[2])
		    + (quat_old[3] + quat_new[3])*(quat_old[3] + quat_new[3])); 

    if (dot2 < dot1) {
	    quat_new[0] = -quat_new[0];
	    quat_new[1] = -quat_new[1];
	    quat_new[2] = -quat_new[2];
	    quat_new[3] = -quat_new[3];
    }

    HC_Open_Segment(old_seg);
    // now loop throw the transition 20 times 
	for (int q=1; q<=num_of_transitions; q++)
	{
		HUtility::TransitionQuaternion(quat_old, quat_new, q/20.0f, quat_interim);
		HUtility::QuaternionToMatrix(quat_interim, matrix_interim);

		// linearly interpolate the scaling factor to the object
		float interim_scale = old_scale + (new_scale - old_scale)*q/20;

		matrix_interim[0] *= interim_scale;
		matrix_interim[1] *= interim_scale;
		matrix_interim[2] *= interim_scale;
		matrix_interim[4] *= interim_scale;
		matrix_interim[5] *= interim_scale;
		matrix_interim[6] *= interim_scale;
		matrix_interim[8] *= interim_scale;
		matrix_interim[9] *= interim_scale;
		matrix_interim[10] *= interim_scale;

		// now that we have a new matrix let's unset the existing matrix and 
		// set the new quaternion matrix on it.
		HC_UnSet_Modelling_Matrix();
		HC_Set_Modelling_Matrix(matrix_interim);
		m_pView->Update();
	}
    HC_Close_Segment();
}


void HUtility::URIdecode(const char * in, char * out, unsigned long * n)
{

    *n=0;

    for(;;){

	char c=*in;
	in++;

	if(c == '\0' || c==' ') 
	    return;

	if(c == '+') 
	    *out++=' ';
	
	else if(c == '%'){
	    unsigned int v=0;
	    bool first = true;

	    while(1){

		    c=*in++;

		    switch((int)c){
			    case 0x0030: v |= 0x0000; break;
			    case 0x0031: v |= 0x0001; break;
			    case 0x0032: v |= 0x0002; break;
			    case 0x0033: v |= 0x0003; break;
			    case 0x0034: v |= 0x0004; break;
			    case 0x0035: v |= 0x0005; break;
			    case 0x0036: v |= 0x0006; break;
			    case 0x0037: v |= 0x0007; break;
			    case 0x0038: v |= 0x0008; break;
			    case 0x0039: v |= 0x0009; break;
			    case 0x0041: v |= 0x000A; break;
			    case 0x0042: v |= 0x000B; break;
			    case 0x0043: v |= 0x000C; break;
			    case 0x0044: v |= 0x000D; break;
			    case 0x0045: v |= 0x000E; break;
			    case 0x0046: v |= 0x000F; break;
			    default: return;
		    }

		    if(first){
			    v <<= 4;
			    first=false;
		    }else 
			    break;
	    }
	    //another possibility for the above while loop would be the following block
	    /*
	    if( *in++ <= 0x39 ) {
		v |= (c - 0x39) << 4;
	    }
	    else {
		v |= (c - 0x41 + 0xA) << 4;
	    } 
	    if( *in++ <= 0x39 ) {
		v |= c - 0x39;
	    }
	    else {
		v |= c - 0x41 + 0xA;
	    }
	    */

	    *out++=(char)v;
	}
	else{
	    *out++ = c;
	}
		++*n;
    }
}

// protected characters: a--z, A--Z, 0--9, *_-.@
// blank is converted into +
// rest of non-protected char becomes %hex
void HUtility::URIencode(const char * in, unsigned long n, char * out)
{
    unsigned int i, j=0;

    for(i=0;i<n;i++) {

	    char c= (char)in[i];

	    if( c == ' ') {
		    out[j] = '+';
		    j++;

	    }
	    else if(c=='_' || c=='-' || c=='*' || c=='.' || c=='@'
		    || c>='a' && c<='z' || c>='A' && c<='Z' || c>='0' && c<='9' ){

		    out[j] = c;
		    j++;

	    }
	    else
	    {
			const char * HEX = "0123456789ABCDEF";

			out[j] = '%';
			j++;
			out[j]=HEX[(((unsigned char)c)>>4)];
			j++;
			out[j]=HEX[(((unsigned char)c)&0x000F)];
			j++;
	    }
    }
    out[j] = '\0';
}



// gets "pushed" attribute setting from open segment
// internal use only
bool HUtility::RestoreAttributeFromOption(const char *attName, char *attText)
{
	char optionText[4096];
	char extAttName[512];
	sprintf(extAttName,"att %s", attName);
  	HC_Show_User_Options(optionText);
	if (strstr(optionText, extAttName))
	{
 		HC_Show_One_User_Option(extAttName,attText);
		HC_UnSet_One_User_Option(extAttName);
		ReplaceCharacter(attText, '$',',');
		return true;
	}
	return false;
}


//is there a string.h function that does this???
void HUtility::ReplaceCharacter(char *text, char orig, char target)
{
	for (unsigned int i=0;i<strlen(text);i++)
		if (text[i] == orig)
			text[i] = target;
}


//calls appropriate HC_Set... function based on attribute name
bool HUtility::SetAttribute(const char *attName, const char *attText)
{
	bool ret = true;
	if (strcmp(attName,"color") == 0)
		HC_Set_Color(attText);
	else if (strcmp(attName,"visibility") == 0)
		HC_Set_Visibility(attText);
	else if (strcmp(attName,"rendering options") == 0)
		HC_Set_Rendering_Options(attText);
	else if (strcmp(attName,"edge weight") == 0)
		HC_Set_Edge_Weight(atof(attText));
	else
		ret = false;
	return (ret);
}


//calls approriate HC_Show... function based on attribute name
//non string values need to be converted
bool HUtility::ShowAttribute(const char *attName, char *attText)
{
	bool ret = true;
	if (strcmp(attName,"color") == 0)
		HC_Show_Color(attText);
	else if (strcmp(attName,"visibility") == 0)
		HC_Show_Visibility(attText);
	else if (strcmp(attName,"rendering options") == 0)
		HC_Show_Rendering_Options(attText);
	else if (strcmp(attName,"edge weight") == 0)
	{
		float weight;
		HC_Show_Edge_Weight(&weight);
		sprintf(attText,"%f",weight);
	}
	else 
		ret = false;

	return (ret);
}

  
/* 
   Stores a local attribute as a user option in the currently open segment for later retrievel
   Useful for highlighting, selection...

  \param attName name of the attribute that needs to be stored (see above for currently supported attributes)
  \return true if local attribute exists

*/
bool HUtility::PushAttribute(const char *attName)
{
	char optionText[4096];
	char attText[4096];
 	if (HC_Show_Existence(attName))
	{
		if (ShowAttribute(attName, attText))
		{		
			sprintf(optionText,"att %s = %s",attName, attText);
			ReplaceCharacter(optionText, ',','$');
			HC_Set_User_Options(optionText);
			return true;
		}
	}
	return false; 
}

/* 
   Stores a local attribute as a user option in the currently open segment for later retrievel
   This function takes the attribute as a string instead of getting it from the segment directly
   Usefull for highlighting, selection...	
   \param attName name of the attribute that needs to be stored (see above for currently supported attributes)
   \param attText Attribute definition (e.g. "faces = red, edges = blue")

*/
void HUtility::PushAttribute(const char *attName, const char *attText)
{
	char optionText[4096];
	
	sprintf(optionText,"att %s = %s",attName, attText);
	ReplaceCharacter(optionText, ',','$');
	HC_Set_User_Options(optionText);
}
 

/* 
   Replaces the requested attribute in the currently open segment with the stored attribute
   The stored user option associated witht that attribute will be deleted	
   \param attName name of the attribute that needs to be stored (see above for currently supported attributes)
   \return true if stored attribute exists

*/
bool HUtility::PopAttribute(const char *attName)
{
	char attText[4096];
	if (HC_Show_Existence("user options"))
	{
		if (RestoreAttributeFromOption(attName, attText))
		{
 			if (SetAttribute(attName, attText))
				return true;
		}
	}
	return false; 
  
}

/* 
   Modifies the stored attribute in the currently open segment
   Example: stored attribute is "faces = red"
   newStyle = "lines = yellow"
   result: "faces = red, lines = yellow"
    \param attName name of the stored attribute that needs to be modified 
   \param modification string 
   \return true if stored attribute exists

*/
bool HUtility::ModifyPushedAttribute(const char *attName, const char *newStyle)
{
	char attText[4096];
	char optionText[4096];
	if (HC_Show_Existence("user options"))
	{
		if (RestoreAttributeFromOption(attName, attText))
		{
			HC_Open_Segment("/att_temp");
			SetAttribute(attName,attText);
			SetAttribute(attName,newStyle);
			ShowAttribute(attName, attText);
			HC_Close_Segment();
			HC_Delete_Segment("/att_temp");
			sprintf(optionText,"att %s = %s",attName, attText);
			ReplaceCharacter(optionText, ',','$');
			HC_Set_User_Options(optionText);
			return true;
			
		}
	}
	return false;
}

bool HUtility::ShowPushedAttribute(const char *attName, char *attText)
{
 	if (HC_Show_Existence("user options"))
	{
		if (RestoreAttributeFromOption(attName, attText))
			return true;
	}
	return false;
}

#define MAX_SEGMENT_NAME_LENGTH		1024
#define MAX_TYPE_NAME_LENGTH		32

/* 
 Extracts all of the shells from the currently open segment and its children, and merges them into one.
*/
HC_KEY HUtility::Merge_Shells( bool do_recurse)
{
    HC_KEY	key;
    HVector	*v;
    int		i;
    HShell	total_info = { 0, NULL, NULL, 0, NULL };
    HShell	max_info   = { 0, NULL, NULL, 0, NULL };
    HShell	optimized  = { 0, NULL, NULL, 0, NULL };
    int		*point_mapping_list, *face_mapping_list;
    
    Segment_Shell_Allocate (&total_info, &max_info, do_recurse);
    Recursive_Merge_Shells (&total_info, &max_info, do_recurse);
    
    optimized.point_list = (HPoint *) 
	    malloc (total_info.point_list_len * sizeof (HPoint));
    optimized.face_list =  (int *) 
	    malloc (total_info.face_list_len * sizeof (int));
    
    point_mapping_list = (int *) 
	    malloc (total_info.point_list_len * sizeof (int));
    face_mapping_list = (int *) 
	    malloc (total_info.face_list_len * sizeof (int));

	// clean out all the existing shells
	HC_Flush_Contents(".", "shells");
    
    /* optimize it */
    HC_Compute_Optimized_Shell ( total_info.point_list_len, total_info.point_list,
					 (float *)total_info.normal_list,
					 total_info.face_list_len, total_info.face_list,
					 "tolerance = 0.0oru, normal tolerance = 0degrees",
					 &(optimized.point_list_len),optimized.point_list,
					 &(optimized.face_list_len), optimized.face_list,
					 point_mapping_list, face_mapping_list );

    key = HC_KInsert_Shell ( optimized.point_list_len, 
				     optimized.point_list,
				     optimized.face_list_len, 
				     optimized.face_list);
    
    HC_Open_Geometry( key );
    for( i = 0 ; i < total_info.point_list_len ; i++ )
    {
	if( point_mapping_list[i] >= 0 )
	{
	    v = &(total_info.normal_list[ i ]);
	    HC_Open_Vertex( point_mapping_list[i] );
	    HC_Set_Normal( v -> x,  v -> y,  v -> z );
	    HC_Close_Vertex();
	}
    }
    HC_Close_Geometry();
    
    if(max_info.point_list)		
	free(max_info.point_list);
    if(max_info.face_list)		
	free(max_info.face_list);
    if(max_info.face_list)		
	free(max_info.normal_list);
    if(total_info.point_list)	
	free(total_info.point_list);
    if(total_info.face_list)	
	free(total_info.face_list);
    if(total_info.normal_list)	
	free(total_info.normal_list);
    if(optimized.point_list)	
	free(optimized.point_list);
    if(optimized.face_list)		
	free(optimized.face_list);
    if(optimized.normal_list)	
	free(optimized.normal_list);
    if(point_mapping_list)		
	free(point_mapping_list);
    if(face_mapping_list)		
	free(face_mapping_list);

    return key;   
}

/* 
 Designed to be called from Segment_Allocate. Finds how much memory necessary for the current 
 and child segments' shells. 
 \param total_info should be big enough to hold all the data for all of the shells put together
 \param max_info should be big enough for the largest shell.
*/
void HUtility::Shell_Search( HShell *total_info, HShell *max_info, bool do_recurse )
{
    int		local_point_list_len;
    int		local_face_list_len;
    HC_KEY	key;
    int		num_shells = 0;
    char	type[ MAX_TYPE_NAME_LENGTH ];
    char	child[MAX_SEGMENT_NAME_LENGTH];
    
    HC_Begin_Segment_Search ("*");
    if (do_recurse)	
    {
	while (HC_Find_Segment (child) ) 
	    {
		HC_Open_Segment (child);
		Shell_Search (total_info, max_info, do_recurse);
		HC_Close_Segment();
	    }
    }
    HC_End_Segment_Search();
    
    HC_Begin_Contents_Search (".", "shells");
    HC_Show_Contents_Count (&num_shells);
    while (HC_Find_Contents (type, &key)) 
    {
	HC_Show_Shell_Size(key, &local_point_list_len, &local_face_list_len );
	/* keep track of the totals  */
	total_info -> point_list_len += local_point_list_len;
	total_info -> face_list_len  += local_face_list_len;
	/* keep track of the maximums */
	if( max_info -> point_list_len < local_point_list_len )
		max_info -> point_list_len = local_point_list_len;
	if( max_info -> face_list_len < local_face_list_len )
		max_info -> face_list_len = local_face_list_len;
    }
    HC_End_Contents_Search ();
	
}

/* 
 Allocates memory for merging the shells contained below the segment open in the current segment tree
 \param total_info after function is called this will be big enough to hold all the shell data 
 \param max_info after function is called this will be big enough to hold the largest shell data
*/
bool HUtility::Segment_Shell_Allocate( HShell *total_info, HShell *max_info, bool do_recurse )
{
    Shell_Search (total_info, max_info, do_recurse);
    
    max_info -> point_list = (HPoint *) 
	    malloc ( max_info -> point_list_len * sizeof (HPoint) );
    max_info -> normal_list = (HVector *)
	    malloc ( max_info -> point_list_len * sizeof (HVector) );
    max_info -> face_list  = (int *)   
	    malloc ( max_info -> face_list_len * sizeof (int) );
    total_info -> point_list = (HPoint *) 
	    malloc ( total_info -> point_list_len * sizeof (HPoint) );
    total_info -> normal_list = (HVector *)
	    malloc ( total_info -> point_list_len * sizeof (HVector) );
    total_info -> face_list  = (int *)   
	    malloc ( total_info -> face_list_len * sizeof (int) );
    
    /* clear the lengths of the structures */
    max_info -> point_list_len	    = 0;
    max_info -> face_list_len	    = 0;
    total_info -> point_list_len    = 0;
    total_info -> face_list_len	    = 0;
    
    /* return 0 if malloc returned NULL for any of the memory requests */
    if((max_info -> point_list) &&
       (max_info -> face_list) &&
       (total_info -> point_list) &&
       (total_info -> face_list))
	    return true;
    else
	    return false;
	
}

/* 
 This function assumes that the proper amount of memory has been allocated (see Segment_Shell_Allocate).
 Recursive_Merge steps through the child segments, and merges all of the shells into one.
 \param total_info merged shell will be contained at this address 
 \param max_info largest shell data will be contained at this address
 \param do_recurse whether or not to only merge shells in the currently open segment or to recurse through
 the segment tree beneath it.
*/
HC_KEY HUtility::Recursive_Merge_Shells( HShell *total_info, HShell *max_info, bool do_recurse, bool delete_orig_shells, bool insert_null_lods  )
{
    int		local_point_list_len;
    int		local_face_list_len;
    int		polygon_length;
    int		i, j;
    char	type[ MAX_TYPE_NAME_LENGTH ];
    HC_KEY	key;
    HVector	*v;
    int		current_point_list_offset, current_face_list_offset;

    HC_KEY  last_shell = -1;
    HC_KEY  tmp;

    /* recursively find all of the child segments */

    HC_Begin_Contents_Search (".", "segment, includes");
    if (do_recurse) 
    {
	    while (HC_Find_Contents (type, &key) ) 
	    {
			if (strstr("include", type))
			{
			    long new_key = HC_KShow_Include_Segment(key);
				HC_Open_Segment_By_Key(new_key);
			}
			else
			    HC_Open_Segment_By_Key(key);

		    tmp = Recursive_Merge_Shells (total_info, max_info, do_recurse  );
		    if (tmp != -1)
				last_shell = tmp;
		    HC_Close_Segment ();
	    }
    }
    HC_End_Contents_Search ();

    /*
     * after all of the children have been processed, process  current segment
     *   (i.e., a post-order tree traversal)
     */
    HC_Begin_Contents_Search (".", "shells");
    while (HC_Find_Contents (type, &key)) 
    {
	last_shell = key;

	current_point_list_offset = total_info -> point_list_len;
	current_face_list_offset  = total_info -> face_list_len;
    
	/*
	 * use max_info for the scratch-space buffers, but
	 *  write the lengths into local variables, just in case 
	 *  we need the max lengths later on.
	 */
	HC_Show_Shell ( key, 
			    &(local_point_list_len), 
			    max_info -> point_list, 
			    &(local_face_list_len), 
			    max_info -> face_list 
			    );

	if( insert_null_lods )
	{
		HC_Open_Geometry(key);
			HC_Open_LOD(1);
				HC_Insert_Shell(0,NULL,0,NULL);
			HC_Close_LOD();
		HC_Close_Geometry();
	}

	/* 
	 * Copy the point and face list into the main face list
	 * For each individual shell, the point list originally started 
	 *   at zero, but for the main list, it starts at 
	 *   current_point_list_offset.  
	 */
	/* here, we copy the point list */
	for( i = 0 ; i < local_point_list_len ; i++ )
	{
	    total_info -> point_list[ current_point_list_offset + i].x = 
		    max_info -> point_list[i].x;
	    total_info -> point_list[ current_point_list_offset + i].y = 
		    max_info -> point_list[i].y;
	    total_info -> point_list[ current_point_list_offset + i].z = 
		    max_info -> point_list[i].z;
	}
    
	/* here, we copy the face list */
	i = 0;
	while(i < local_face_list_len)
	{
	    /*
	     * in this loop:
	     * main_face_list[current_face_list_offset]  
	     *      is the start of the individual shell
	     * main_face_list[current_face_list_offset + i]  
	     *     is the start of a polygon inside the shell
	     * main_face_list[current_face_list_offset + i + j]
	     *      is a point inside the polygon inside the shell
	     * the face list is made up of polygon entries.  For each entry,
	     *      the length comes first, then a number of data points
	     *      equal to the length.
	     */
	    polygon_length = max_info -> face_list[i];
	    /* 
	     * hoops syntax: to subtract a polygon from a shell, use negative
	     *      polygon length.  We account for this here with abs. val.
	     */
	    if( polygon_length < 0 ) polygon_length = -polygon_length;
	    total_info -> face_list[ i + current_face_list_offset ] =  
		    max_info -> face_list[i];
	    for( j = 1 ; j <= polygon_length ; j++ ) 
	    {
		    /*
		     * remember, the point list was moved, so the face list 
		     * has to be modified as it is copied
		     */
		    total_info -> face_list[ current_face_list_offset + i + j ] = 
			    max_info -> face_list[ i + j ] + current_point_list_offset;
	    }
	    i += polygon_length + 1;
	    
	} /* end while ( i < local_face_list_len ) */
    
	HC_Open_Geometry( key );
	for( i = 0 ; i < local_point_list_len ; i++ )
	{
	    v = &(total_info -> normal_list[current_point_list_offset+i]);
	    HC_Open_Vertex( i );
	    HC_Show_Net_Normal( &(v -> x), &(v -> y), &(v -> z) );
	    HC_Close_Vertex();
	}
	HC_Close_Geometry();
    
	/* update the current offsets and prepare to load the next shell. */
	total_info -> point_list_len += local_point_list_len;
	total_info -> face_list_len  += local_face_list_len;
	    
    } /* end while ( HC_Find_Contents(...) ) */
    HC_End_Contents_Search ();


	if( delete_orig_shells )
		HC_Flush_Contents (".", "shells");

    return last_shell;
}


/* 
 This function merges all the shells in the segement and generates LODs for the merged shell. These
 LODs are then attached to one of the shells in the segment.
 \param do_recurse segment will be recursively searched for shells
 \param delete_orig_shells once the shells are merged, the orginal shells will be erased and merged shell will remain
 \param insert_null_lods null shells will be attached as LOD geometry for all the shells except one
 \param n_lod_levels no of lod levels to be generated
 \param reduction_ratio ratio of no of triangles in LOD vs. the no of triangles in previous LOD
*/
void HUtility::Generate_LODs( bool do_recurse, bool delete_orig_shells, bool insert_null_lods, 
							 int n_lod_levels, float reduction_ratio )
{
    HShell	total_info = { 0, NULL, NULL, 0, NULL };
    HShell	max_info   = { 0, NULL, NULL, 0, NULL };
    
    Segment_Shell_Allocate (&total_info, &max_info, do_recurse);
    HC_KEY shellkey = Recursive_Merge_Shells (&total_info, &max_info, do_recurse, delete_orig_shells, insert_null_lods);
    
	char lod_ratio[256];
	sprintf(lod_ratio,"lod = %f%%",reduction_ratio*100.0);

    for (int count = 1; count <= n_lod_levels; count++)
    {
	// let's only do this if we have a good sized merged shell
		if (total_info.point_list_len > 100)
		{

			// reduce each LOD to ratio% of the vertex size of the previous LOD
			HC_Compute_Optimized_Shell ( total_info.point_list_len, total_info.point_list,
							 NULL,
							 total_info.face_list_len, total_info.face_list,
							 lod_ratio,
							 &(total_info.point_list_len), total_info.point_list,
							 &(total_info.face_list_len), total_info.face_list,
							 NULL, NULL );

    
			HC_Open_Geometry(shellkey);
			HC_Open_LOD(count);
				HC_Insert_Shell ( total_info.point_list_len, 
							 total_info.point_list,
							 total_info.face_list_len, 
							 total_info.face_list);
			HC_Close_LOD(); 
			HC_Close_Geometry();
		}
    }

    if(max_info.point_list)		
	free(max_info.point_list);
    if(max_info.face_list)		
	free(max_info.face_list);
    if(max_info.face_list)		
	free(max_info.normal_list);
    if(total_info.point_list)	
	free(total_info.point_list);
    if(total_info.face_list)	
	free(total_info.face_list);
    if(total_info.normal_list)	
	free(total_info.normal_list);

    return;
}

/*!
	Counts the number of faces, vertices and triangles in a given shell primitive. 
	\param shell_key HOOPS key to a given shell  
	\param faces Number of faces in the given shell.  Passed by reference.
	\param vertices Number of vertices in the shell, including duplicates.  Passed by reference
	\param triangles Number of triangles in the shell after HOOPS triangulation calculation.  Passed by reference.
*/
void HUtility::CountShellData (
    HC_KEY      shell_key,
    int *       faces,
    int *       vertices,
    int *       triangles,
    int *	tristrip_count) {

    int     i, j, k, ts_length, numtriangles, numvertices;
    int	    numfaces = 0, numtristrips = 0;
    int	    *tristrips, *face_indices, *iend, *iptr;
    bool    *used; //tracks uniqueness so each face counts only once towards *faces output

    /* show by tristrips because if the model was inserted by tristrips, we don't want
     * to have to compute a dummy standard-format face list just for this computation. */
    HC_Show_Shell_By_Tristrips_Size (shell_key, &numvertices, &ts_length, &numtriangles);

    tristrips  = new int[ ts_length ];
    face_indices = new int[ numtriangles ];
    used = new bool[ numtriangles ];

    HC_Show_Shell_By_Tristrips (shell_key, 0, 0, 0, tristrips, 0, face_indices);
    for( i = 0 ; i < numtriangles ; i++ ) {
        used[i] = false;
    }

    /* find the number of faces */
    k = 0;
    for( i = 0 ; i < ts_length ; i += tristrips[i] + 1 ) {
        for( j = 2 ; j < tristrips[i] ; j++ ) {
            if( !used[ face_indices[k] ] ) {
                used[ face_indices[k] ] = true;
                numfaces++;
            }
            k++;
        }
    }

    /* find the number of tristrips */
    iend = tristrips + ts_length;
    iptr = tristrips;
    while ( iptr < iend ) {
	if( *iptr > 0 ) 
	    iptr += *iptr + 1;
	else
	    iptr += -(*iptr) + 1;
	numtristrips++;
    }

    delete [] tristrips;
    delete [] face_indices;
    delete [] used;

    if( faces != 0 )
        *faces = numfaces;
    if( triangles != 0 )
        *triangles = numtriangles;
    if( vertices != 0 )
        *vertices = numvertices;
    if( tristrip_count != 0 )
        *tristrip_count = numtristrips;
}


void HUtility::FindFileNameExtension(const char * filename_in, char * ext_out)
{
	ext_out[0] = '\0';

	HC_Parse_String(filename_in, ".", -1, ext_out);

	// make it lower case
	unsigned int i;
	for(i = 0; i < strlen(ext_out); i++)
		ext_out[i] = (char) tolower(ext_out[i]);
}

void HUtility::FindFileNameAndDirectory(const char * filename_in, char * dir_out, char * filename_out)
{

	strcpy(dir_out, filename_in);

	unsigned int i = strlen(filename_in);

	while( i!=0 && dir_out[i] != '/' && dir_out[i] != '\\') i--;

	if(i!=0) {

		//SKIP MULT SEPERATORS
		while (dir_out[i] != '\0' && (dir_out[i] == '/' || dir_out[i] == '\\'))i++;

		if(dir_out[i] == '\0') {
			//NO FILE
			filename_out[0] = '\0';
		} else {
			strcpy(filename_out, (const char *)(&dir_out[i]));
			dir_out[i]='\0';
		}

	}else{
		//NO DIR
		strcpy(filename_out, filename_in);
		dir_out[i]='\0';
	}

}






