//
// Copyright (c) 2000 by Tech Soft 3D, LLC.
// The information contained herein is confidential and proprietary to
// Tech Soft 3D, LLC., and considered a trade secret as defined under
// civil and criminal statutes.  Tech Soft 3D shall pursue its civil
// and criminal remedies in the event of unauthorized use or misappropriation
// of its trade secrets.  Use of this information by anyone other than
// authorized employees of Tech Soft 3D, LLC. is granted only under a
// written non-disclosure agreement, expressly prescribing the scope and
// manner of such use.
//
// $Header: /files/homes/master/cvs/hoops_master/docs_hoops/HoopsMFC/CHoopsView.h,v 1.2 2006-08-07 20:39:02 stage Exp $
//

// CHoopsView.h : interface of the CHoopsView class, derived from CView
// Adds encapsulation of a HOOPS View Object, which corresponds to an
// instance of a HOOPS driver
//
// Also provides high-level support for copying the view to the clipboard, 
// printing and printpreview

#ifndef hoops_view
#define hoops_view


#include "HBaseView.h"

class CHoopsDoc;

// msw driver specific driver_option debug flags
#define DEBUG_NO_WINDOWS_HOOK			0x00000040
#define DEBUG_FORCE_FULL_COLOR			0x00000010
#define DEBUG_SFB_COPY_TO_CLIPBOARD		0x00004000
#define DEBUG_NO_RASTERS_IN_PRINTING	0x00020000
#define DEBUG_PRINTING_MODE				0x00040000
#define DEBUG_CLIPBOARD_MODE			0x00080000

// OpenGL driver specific debug flags
#define DEBUG_16BIT_ZBUFFER				0x00200000

#define WMF             0
#define EMF             1

// generic rendering options debug flags
#define DEBUG_NO_PAINTERS_PANELLING	0x00000008


//! The CHoopsView class provides a HOOPS-specific implementation of the CView object.
/*!
	CHoopsView creates and manages the connection of a HOOPS/3dGS driver instance to a CView object, which includes
	properly updating the window during exposes/resizes.  It contains a pointer to a corresponding HOOPS/MVO HBaseView object.

	It includes built-in support for printing, print preview, and copying to cliboard of the HOOPS/3dGS scene, 
	and encapsulates palette management if HOOPS/3dGS is being used on 8-bit system with the MSW Driver. 

	This is the base class from which all custom HOOPS-based MFC/MDI CView objects should be derived.
*/
class AFX_EXT_CLASS CHoopsView : public CView
{
private:

	bool			m_first_update;			// flag indicating whether the first update has occurred (meaning the window is mapped, etc...)

protected: // create from serialization only
	CHoopsView();

	DECLARE_DYNCREATE(CHoopsView);


    CPalette		*m_pViewPalette;  	/*!< Pointer to this view's unique palette. This will be null if we're in 24-bit mode, or if palettes are shared among all views. */

	/*! 
		If true, the scene is rendered to an offscreen bitmap and stretch-blitted to the printer or Windows metafile 
		during printing or copy to clipboard. The default is true.
	*/
    bool			m_bFastPrint;   
	
    /*! 
		If true, the scene is rendered to a windows metafile and then played to the 
		printer. The default is false.
	*/
	bool			m_bMetaPrint;      
	
    /*! If true, the scene will render to the clipboard in truecolor mode. The default is true.*/
	bool			m_bClipboardTruecolor;

	/*! Type of metafile generated when m_bMetaPrint==true, 0=WMF amd 1=EMF.*/
    int				m_MetafileType;

	/*! Pointer to HBaseView MVO object associated with this CView object.  */
	HBaseView*		m_pHView;		

	/*! 
	  The key of the HOOPS driver instance segment associated with this view
	  Intended to be used as an ALTERNATIVE to MVO (and the HBaseView object referenced by m_pHView)
	*/
	long			m_ViewKey;	
	
	/*! 
	  The key of the HOOPS segment representing the top of the scene hierarchy, typically a subsegment of the driver instance 
	  segment (m_ViewKey)   Intended to be used in conjuction with m_view_key as an ALTERNATIVE to MVO (and the HBaseView 
	  object referenced by m_pHView)
	*/
	long			m_SceneKey;	
	
	unsigned long MapFlags( unsigned long state ); //map to mvo codes						

// Operations
public:
	CHoopsDoc* GetDocument();

	/*! \return The pointer to the HBaseView object associated with this CView object. */
	HBaseView* GetHoopsView( void ) { return m_pHView; }


    CPalette*   GetPalette();

    void        InitializePalette();

	// printing and metafile methods

	/*! 
		\return A Boolean indicating whether native GDI printing or a fast bitmap blit will be used for printing.
				The default is true.
	*/
    bool                GetFastPrint (void) {return m_bFastPrint;};

	/*! Sets whether native GDI printing or a fast bitmap blit will be used for printing. The default is true.*/
    void                SetFastPrint (bool fastprint) {m_bFastPrint = fastprint;};

	/*! 
		\return A Boolean indicating whether a Windows Enhanced Metafile will first be created and then played to 
		the printer when CHoopsView::OnPrint() is called. The default is false.
	*/
	bool                GetMetaPrint (void) {return m_bMetaPrint;};

	/*! Sets whether a Windows Enhanced Metafile containing the scene will first be created and then played to 
		the printer when CHoopsView::OnPrint() is called (rather than the scene going straight to the printer). The default is false.
	*/
    void                SetMetaPrint (bool metaprint) {m_bMetaPrint = metaprint;};

	/*! \return An integer indicating whether an emf or wmf will be for clipboard rendering.  The default is EMF. */
    int			        GetMetafileType (void) { return m_MetafileType; };

	/*! Sets whether an emf or wmf will be used for clipboard rendering.  The default is EMF. */
    void		        SetMetafileType (int emf_or_wmf) { m_MetafileType = emf_or_wmf; };

	/*! \return A Boolean indicating whether the first update has occurred. */
    bool				GetFirstUpdate(){return m_first_update;};

	/*! Sets whether the first view update has occurred.*/
    void				SetFirstUpdate(bool firstupdate){m_first_update = firstupdate;};

	/*! \return A Boolean indicating whether copying to clipboard uses mapped or true color. The default is true. */
	bool		        GetClipboardTruecolor (void) { return m_bClipboardTruecolor; };

	/*! Sets whether copying to clipboard uses mapped or true color.  The default is true. */
    void		        SetClipboardTruecolor (bool truecolor) { m_bClipboardTruecolor = truecolor; };


	void		SaveEMF(char *filename); 


	long		GetViewKey();
	long		GetSceneKey();

	void		AdjustAxisWindow();

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CHoopsView)
	public:
	virtual void OnDraw(CDC* pDC);  // overridden to draw this view
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnInitialUpdate();
	virtual BOOL Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext = NULL);
	protected:
	virtual BOOL OnPreparePrinting(CPrintInfo* pInfo);
	virtual void OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnEndPrinting(CDC* pDC, CPrintInfo* pInfo);
	virtual void OnPrint(CDC* pDC, CPrintInfo* pInfo);
 
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CHoopsView();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CHoopsView)
	afx_msg void OnPaint();
	afx_msg void OnEditCopy();
	afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags);
	afx_msg void OnLButtonDblClk(UINT nFlags, CPoint point);
	afx_msg virtual void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg virtual void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDblClk(UINT nFlags, CPoint point);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnRButtonUp(UINT nFlags, CPoint point);
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message);
	afx_msg virtual void OnTimer(UINT nIDEvent);
	afx_msg void OnSize( UINT nType, int cx, int cy );
	afx_msg void OnWindowPosChanged( WINDOWPOS* lpwndpos );

	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

#ifndef _DEBUG  
inline CHoopsDoc* CHoopsView::GetDocument()
   { return (CHoopsDoc*)m_pDocument; }
#endif

#endif

/////////////////////////////////////////////////////////////////////////////
